﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ClientSide;
using SharedPacketLib.DataPackets;
using System.Threading;
using SharedPacketLib.Data_Format;

namespace ClientImplementation.ConnectionClasses
{
    public class ClientConnectionManager : IDisposable
    {
        /// <summary>
        /// The packet register 
        /// </summary>
        private PacketProcessor processor;

        /// <summary>
        /// Contains the connection information about the item
        /// </summary>
        private ClientSocketInformation connection;

        /// <summary>
        /// Indicates if this item 
        /// </summary>
        private bool isConnecting;

        public delegate void connectionChanged(ConnectionState state);

        /// <summary>
        /// If the connection state is changed this event is triggered
        /// </summary>
        public event connectionChanged OnConnectionChange;

        private string ip;
        private int port;

        /// <summary>
        /// Creates a new connection manager of thisitem
        /// </summary>
        /// <param name="ip">The ip to connect to</param>
        /// <param name="port">The port that should be contacted</param>
        public ClientConnectionManager(string ip, int port)
        {
            this.ip = ip;
            this.port = port;
            connection = new ClientSocketInformation(ip, port, true);
            connection.connectionChanged += reportNewConnectionState;
            connection.packetArrival += packetArrival;
            processor = new PacketProcessor();
        }

        /// <summary>
        /// Resets the connection and cleans the packet register
        /// </summary>
        public void reset()
        {
            connection.Dispose();
            connection.connectionChanged -= reportNewConnectionState;
            connection.packetArrival -= packetArrival;

            connection = new ClientSocketInformation(ip, port, true);
            connection.connectionChanged += reportNewConnectionState;
            connection.packetArrival += packetArrival;
            processor = new PacketProcessor();
        }

        /// <summary>
        /// Adds or replaces a packet handeler
        /// </summary>
        /// <param name="code">The Client code</param>
        /// <param name="packetFunction">The packet function which can handle a packet</param>
        public void registerPacket(ClientOpCode code, PacketRegistery<ClientOpCode, ClientIncomingPacket>.handlePacket packetFunction, bool handleImmidiate = false)
        {
            this.processor.registerPacket(code, packetFunction, handleImmidiate);
        }

        /// <summary>
        /// Removes a packet from the opcode list
        /// </summary>
        /// <param name="code">The opcode to remove</param>
        public void removePacket(ClientOpCode code)
        {
            this.processor.removePacket(code);
        }

        /// <summary>
        /// Called when a packet has been received
        /// </summary>
        /// <param name="packet"></param>
        private void packetArrival(ClientIncomingPacket packet)
        {
            processor.handlePacket(packet);
        }

        /// <summary>
        /// Opens a new connection and triggers the event
        /// </summary>
        public void openConnection()
        {
            if (connection.isConnectedToServer())
            {
                reportNewConnectionState(ConnectionState.open);
            }
            else if (!isConnecting)
            {
                isConnecting = true;
                Thread t = new Thread(new ThreadStart(connectToServer));
                t.Start();
            }
        }

        /// <summary>
        /// Connects this item to the server
        /// </summary>
        private void connectToServer()
        {
            connection.ConnectToServer();
            isConnecting = false;
        }

        /// <summary>
        /// Reports the new connection state of the connection
        /// </summary>
        private void reportNewConnectionState(ConnectionState state)
        {
            if (this.OnConnectionChange != null)
                this.OnConnectionChange.Invoke(state);
        }

        /// <summary>
        /// Sends a packet to the server if its connected
        /// </summary>
        /// <param name="packet"></param>
        public void sendData(ISendable packet)
        {
            connection.sendData(packet.getPacket() as ClientOutgoingPacket);
        }

        /// <summary>
        /// Processes all messages in the since the last time this item was called
        /// </summary>
        public void processSyncedMessages()
        {
            processor.processPackets();
        }

        public void Dispose()
        {
            connection.Dispose();
        }
    }
}
