﻿using System;
using System.Net.Sockets;
using SharedPacketLib.DataPackets;
using SharedPacketLib.Data_Format.Data_Exception;
using System.Net;


namespace ClientSide
{
    public class ClientSocketInformation : IDisposable
    {
        #region declares
        /// <summary>
        /// The socket this connection is based upon
        /// </summary>
        private Socket dataSocket;

        /// <summary>
        /// The ip of this connection
        /// </summary>
        private string ip;
        
        /// <summary>
        /// The port of the connection
        /// </summary>
        private int port;

        /// <summary>
        /// Boolean indicating of this instance is connected to the user or not
        /// </summary>
        private bool isConnected;

        /// <summary>
        /// Is used every time a packet arrives in this connection
        /// </summary>
        /// <param name="packet">The packet of the item</param>
        public delegate void PacketArrival(ClientIncomingPacket packet);

        /// <summary>
        /// Is triggered every time a packet arrived in this location
        /// </summary>
        public event PacketArrival packetArrival;

        /// <summary>
        /// Is used when a connection state changes
        /// </summary>
        /// <param name="state">The new state of the connection</param>
        public delegate void ConnectionChange(ConnectionState state);
        /// <summary>
        /// Is triggered when the user connects/disconnects
        /// </summary>
        public event ConnectionChange connectionChanged;

        /// <summary>
        /// Buffer of the connection
        /// </summary>
        private byte[] buffer;
        private bool disableNaglesAlgorithm;

        #endregion

        #region constructor

        /// <summary>
        /// Creates a new Connection witht he given information
        /// </summary>
        /// <param name="dataStream">The Socket of the connection</param>
        /// <param name="connectionID">The id of the connection</param>
        public ClientSocketInformation(string address, int port, bool disableNaglesAlgorithm)
        {
            this.buffer = new byte[ConnectionSettings.BUFFER_SIZE];
            this.disableNaglesAlgorithm = disableNaglesAlgorithm;
            this.ip = address;
            this.port = port;
            this.isConnected = false;
            this.packetParser = new DataParser();
            this.packetParser.packetArrival += packetParser_packetArrival;
        }

        private void packetParser_packetArrival(ClientIncomingPacket packet)
        {
            if (this.packetArrival != null)
                this.packetArrival.Invoke(packet);
        }

        

        /// <summary>
        /// Starts this item packet processor
        /// MUST be called before sending data
        /// </summary>
        public bool ConnectToServer()
        {
            if (this.connectionChanged != null)
                this.connectionChanged.Invoke(ConnectionState.connecting);
            if (!isConnected)
            {
                try
                {
                    IPAddress address = null;
                    try
                    {
                        address = IPAddress.Parse(this.ip);
                    }
                    catch
                    {
                        IPHostEntry entry = Dns.GetHostEntry(this.ip);
                        for (int i = 0; i < entry.AddressList.Length; i++)
                        {
                            if ((entry.AddressList[i].AddressFamily == AddressFamily.InterNetworkV6) && ConnectionSettings.IPV6_Enabled)
                            {
                                address = entry.AddressList[i];
                                break;
                            }
                            else if((entry.AddressList[i].AddressFamily == AddressFamily.InterNetwork))
                            {
                                address = entry.AddressList[i];
                                break;
                            }
                        }
                        if (address == null)
                        {
                            throw new Exception();
                        }
                    }
                    IPEndPoint ipe = new IPEndPoint(address, port);
                    this.dataSocket = new Socket(ipe.AddressFamily, SocketType.Stream, ProtocolType.Tcp);
                    this.dataSocket.Connect(ipe);
                    this.dataSocket.NoDelay = disableNaglesAlgorithm;
                    this.dataSocket.BeginReceive(buffer, 0, buffer.Length, SocketFlags.None, new AsyncCallback(incomingDataPacket), dataSocket);
                    Console.WriteLine("Starting packet processsing of address [{0}:{1}]", this.ip, this.port);
                    this.isConnected = true;
                    if (this.connectionChanged != null)
                        this.connectionChanged.Invoke(ConnectionState.open);
                    return true;
                }
                catch
                {
                    
                    disconnect();
                    if (this.connectionChanged != null)
                        this.connectionChanged.Invoke(ConnectionState.failed);
                    return false;
                }
            }
            return false;
        }

        #endregion

        #region getters

        /// <summary>
        /// Returns the ip of the current connection
        /// </summary>
        /// <returns>The ip of this connection</returns>
        public string getIp()
        {
            return ip;
        }


        #endregion

        #region methods

        #region connection management
        /// <summary>
        /// Disconnects the current connection
        /// </summary>
        internal void disconnect()
        {
            if (isConnected)
            {
                try
                {
                    this.packetParser.reset();
                    if (this.dataSocket != null && this.dataSocket.Connected)
                    {
                        dataSocket.Close();
                    }
                }
                catch { }
                this.isConnected = false;
                if (connectionChanged != null)
                    connectionChanged.Invoke(ConnectionState.closed);
            }
        }

        /// <summary>
        /// Disposes the current item
        /// </summary>
        public void Dispose()
        {
            if (isConnected)
            {
                this.disconnect();
            }

        }
        #endregion

        #region data receiving

        /// <summary>
        /// Receives a packet of data and processes it
        /// </summary>
        /// <param name="iAr">The interface of an async result</param>
        private void incomingDataPacket(IAsyncResult iAr)
        {
            int bytesReceived;
            try
            {
                //The amount of bytes received in the packet
                bytesReceived = dataSocket.EndReceive(iAr);
            }
            #region socket error / closed
            catch
            {
                disconnect();
                return;
            }
            if (bytesReceived == 0)
            {
                disconnect();
                return;
            }
            #endregion
            try
            {
                byte[] packet = new byte[bytesReceived];
                Array.Copy(buffer, packet, bytesReceived);
                this.packetParser.handlePacketData(packet);
            }
            #region exception handeling and continueing receiving
            catch (PacketMalformedException)
            {
                //TODO create new event for malformed packet
            }

            catch (SocketException)
            {
                this.disconnect();
                return;
            }
            catch (SendableException ex)
            {
                if (ex.getPacket() != null)
                    this.sendData(ex.getPacket() as ClientOutgoingPacket);
            }

            finally
            {
                if (isConnected)
                {
                    try
                    {
                        //and we keep looking for the next packet
                        this.dataSocket.BeginReceive(buffer, 0, buffer.Length, SocketFlags.None, new AsyncCallback(incomingDataPacket), dataSocket);
                    }
                    catch (SocketException) { disconnect(); }
                }
            }
            #endregion
        }


       

        #endregion

        #region data sending
        /// <summary>
        /// Sends data to this instance
        /// </summary>
        /// <param name="Data">The data to be send</param>
        public void sendData(ClientOutgoingPacket data)
        {
            try
            {
                //convert packet to raw byte data
                byte[] dataBytes = data.GetPacketData();
                //send da shit
                this.dataSocket.BeginSend(dataBytes, 0, dataBytes.Length, 0, new AsyncCallback(sentData), this.dataSocket);
            }
            catch
            {
                disconnect();
            }
        }

        /// <summary>
        /// Sends byte data to the server
        /// </summary>
        /// <param name="packet"></param>
        public void sendData(byte[] packet)
        {
            this.dataSocket.BeginSend(packet, 0, packet.Length, 0, new AsyncCallback(sentData), this.dataSocket);
        }

        /// <summary>
        /// Same as sendData
        /// </summary>
        /// <param name="iAr">The a-synchronious interface</param>
        private void sentData(IAsyncResult iAr)
        {
            try { dataSocket.EndSend(iAr); }
            catch { disconnect(); }
        }
        #endregion

        #endregion

        public bool isConnectedToServer()
        {
            return isConnected;
        }

        internal DataParser packetParser { get; set; }
    }
}
