﻿namespace Database_Manager.Database
{
    using Database_Manager.Database.Database_Exceptions;
    using Database_Manager.Database.Session_Details.Interfaces;
    using Database_Manager.Managers.Database;
    using MySql.Data.MySqlClient;
    using System;
    using System.Collections.Generic;
    using System.Data;
    using System.Runtime.CompilerServices;
    using System.Threading;
    using System.Text;

    public class DatabaseManager
    {
        private int beginClientAmount;
        private string connectionString;
        private List<DatabaseClient> databaseClients;
        private bool isConnected = false;
        private uint maxPoolSize;
        private DatabaseServer server;

        public event DatabaseEvent dbEvent;
        private int waitingThreads;
        private int loopingThreads;
        private int loopThreadStatus;

        public DatabaseManager(uint maxPoolSize, int clientAmount)
        {
            if (maxPoolSize < clientAmount)
            {
                throw new DatabaseException("The poolsize can not be larger than the client amount!");
            }
            this.beginClientAmount = clientAmount;
            this.maxPoolSize = maxPoolSize;
        }

        private void addConnection(int id)
        {
            DatabaseClient item = new DatabaseClient(this, id);
            item.connect();
            this.databaseClients.Add(item);
        }

        private void createNewConnectionString()
        {
            MySqlConnectionStringBuilder connectionString = new MySqlConnectionStringBuilder {
                Server = this.server.getHost(),
                Port = this.server.getPort(),
                UserID = this.server.getUsername(),
                Password = this.server.getPassword(),
                Database = this.server.getDatabaseName(),
                MinimumPoolSize = this.maxPoolSize / 2,
                MaximumPoolSize = this.maxPoolSize,
                AllowZeroDateTime = true,
                ConvertZeroDateTime = true
            };
            this.setConnectionString(connectionString);
        }

        public void destroy()
        {
            lock (this)
            {
                this.isConnected = false;
                if (this.databaseClients != null)
                {
                    foreach (DatabaseClient client in this.databaseClients)
                    {
                        if (!client.isAvailable())
                        {
                            client.Dispose();
                        }
                        client.disconnect();
                    }
                    this.databaseClients.Clear();
                }
            }
        }

        private void disconnectUnusedClients()
        {
            lock (this)
            {
                foreach (DatabaseClient client in this.databaseClients)
                {
                    if (client.isAvailable())
                    {
                        client.disconnect();
                    }
                }
            }
        }

        private DatabaseClient getClient(bool autoCommit)
        {
            waitingThreads++;
            lock (this)
            {
                loopThreadStatus = 0;
                waitingThreads--;
                loopingThreads++;
                DateTime now = DateTime.Now;
                for (int i = 0; i < this.databaseClients.Count; i++)
                {
                    loopThreadStatus = 1;
                    if (this.databaseClients[i].isAvailable())
                    {
                        loopThreadStatus = 2;
                        if (this.databaseClients[i].getConnectionState() == ConnectionState.Closed)
                        {
                            loopThreadStatus = 3;
                            this.databaseClients[i].connect();
                            loopThreadStatus = 4;
                        }
                        else if ((this.databaseClients[i].getConnectionState() == ConnectionState.Broken) || ((now - this.databaseClients[i].getLastAction()).TotalHours > 1.0))
                        {
                            loopThreadStatus = 5;
                            this.databaseClients[i].disconnect();
                            loopThreadStatus = 6;
                            this.databaseClients[i] = new DatabaseClient(this, i);
                            loopThreadStatus = 7;
                            this.databaseClients[i].connect();
                            loopThreadStatus = 8;
                        }
                        if (this.databaseClients[i].getConnectionState() == ConnectionState.Open)
                        {
                            loopThreadStatus = 9;
                            this.databaseClients[i].prepare(autoCommit);
                            loopThreadStatus = 10;
                            loopingThreads--;
                            return this.databaseClients[i];
                        }
                    }
                }
                if (this.databaseClients.Count < this.maxPoolSize)
                {
                    loopThreadStatus = 11;
                    this.addConnection(this.databaseClients.Count);
                    loopThreadStatus = 12;
                    return this.getClient(autoCommit);
                }
                loopThreadStatus = 13;
                this.maxPoolSize = (uint) (this.maxPoolSize * 1.5);
                loopThreadStatus = 14;
                this.createNewConnectionString();
                loopThreadStatus = 15;
                return this.getClient(autoCommit);
            }
        }

        internal string getConnectionString()
        {
            return this.connectionString;
        }

        public int getOpenConnectionCount()
        {
            int num = 0;
            lock (this)
            {
                foreach (DatabaseClient client in this.databaseClients)
                {
                    if ((((client.getConnectionState() == ConnectionState.Open) || (client.getConnectionState() == ConnectionState.Fetching)) || (client.getConnectionState() == ConnectionState.Executing)) || (client.getConnectionState() == ConnectionState.Fetching))
                    {
                        num++;
                    }
                }
            }
            return num;
        }

        public IQueryAdapter getQueryreactor()
        {
            return this.getQueryreactor(false);
        }

        public IQueryAdapter getQueryreactor(bool transaction)
        {
            return this.getClient(transaction).getQueryReactor();
        }

        public void init()
        {
            try
            {
                DatabaseClient client;
                MySqlConnectionStringBuilder builder = new MySqlConnectionStringBuilder();
                this.createNewConnectionString();
                this.databaseClients = new List<DatabaseClient>((int) this.maxPoolSize);
                if (this.beginClientAmount != 0)
                {
                    for (int i = 0; i < this.beginClientAmount; i++)
                    {
                        client = new DatabaseClient(this, i);
                        if (this.dbEvent != null)
                        {
                            this.dbEvent(client);
                        }
                        client.connect();
                        this.databaseClients.Add(client);
                    }
                }
                else
                {
                    client = new DatabaseClient(this, -1);
                    client.connect();
                    client.disconnect();
                }
            }
            catch (MySqlException exception)
            {
                this.isConnected = false;
                throw new Exception("Could not connect the clients to the database: " + exception.Message);
            }
            this.isConnected = true;
        }

        public bool isConnectedToDatabase()
        {
            return this.isConnected;
        }

        private void setConnectionString(MySqlConnectionStringBuilder connectionString)
        {
            this.connectionString = connectionString.ToString();
        }

        public bool setServerDetails(string host, uint port, string username, string password, string databaseName)
        {
            try
            {
                this.server = new DatabaseServer(host, port, username, password, databaseName);
                return true;
            }
            catch (DatabaseException)
            {
                this.isConnected = false;
                return false;
            }
        }

        public delegate void DatabaseEvent(DatabaseClient client);

        public void DumpData(StringBuilder builder)
        {
            Console.WriteLine("Waiting threads: " + waitingThreads);
            builder.AppendLine("Waiting threads: " + waitingThreads);
            Console.WriteLine("Looping threads: " + loopingThreads);
            builder.AppendLine("Looping threads: " + loopingThreads);
            Console.WriteLine("Status: " + loopThreadStatus);
            builder.AppendLine("Status: " + loopThreadStatus);

            Console.WriteLine("Trying to lock le db manager");
            builder.AppendLine("Trying to lock le db manager");
            DateTime start = DateTime.Now;
            lock (this)
            {
                TimeSpan spent = DateTime.Now - start;
                builder.AppendLine("Lock success! " + (int)spent.TotalMilliseconds + "ms");
                Console.WriteLine("Lock success! " + (int)spent.TotalMilliseconds + "ms");
                for (int i = 0; i < this.databaseClients.Count; i++)
                {
                    DatabaseClient dbClient = databaseClients[i];
                    builder.AppendLine("Status for connection: " + i);
                    builder.AppendLine("Pending reset: " + dbClient.pendingReset());
                    builder.AppendLine(dbClient.getConnectionState().ToString());
                    builder.AppendLine();

                    Console.WriteLine("Status for connection: " + i);
                    Console.WriteLine("Pending reset: " + dbClient.pendingReset());
                    Console.WriteLine(dbClient.getConnectionState().ToString());
                    Console.WriteLine();
                }
            }

        }
    }
}

