﻿using System;
using System.Net.Sockets;
using Helpers;
using SharedPacketLib;
using SharedPacketLib.Data_Format.Data_Exception;
using SharedPacketLib.DataPackets;


namespace ConnectionManager
{
    public class ConnectionInformation : IDisposable
    {
        #region declares
        /// <summary>
        /// The socket this connection is based upon
        /// </summary>
        private Socket dataSocket;

        /// <summary>
        /// The manager which created this class
        /// </summary>
        private SocketManager manager;

        /// <summary>
        /// The ip of this connection
        /// </summary>
        private string ip;

        /// <summary>
        /// The id of this connection
        /// </summary>
        private int connectionID;

        /// <summary>
        /// Boolean indicating of this instance is connected to the user or not
        /// </summary>
        private bool isConnected;



        /// <summary>
        /// Is used when a connection state changes
        /// </summary>
        /// <param name="state">The new state of the connection</param>
        public delegate void ConnectionChange(ConnectionInformation information, ConnectionState state);
        /// <summary>
        /// Is triggered when the user connects/disconnects
        /// </summary>
        public event ConnectionChange connectionChanged;

        /// <summary>
        /// Buffer of the connection
        /// </summary>
        private byte[] buffer;

        /// <summary>
        /// This item contains the data parser for the connection
        /// </summary>
        public IDataParser parser { get; private set; }


        #endregion

        #region constructor

        /// <summary>
        /// Creates a new Connection witht he given information
        /// </summary>
        /// <param name="dataStream">The Socket of the connection</param>
        /// <param name="connectionID">The id of the connection</param>
        public ConnectionInformation(Socket dataStream, int connectionID, SocketManager manager, IDataParser parser, string ip)
        {
            this.parser = parser;
            this.buffer = new byte[GameSocketManagerStatics.BUFFER_SIZE];
            this.manager = manager;
            this.dataSocket = dataStream;
            this.dataSocket.SendBufferSize = GameSocketManagerStatics.BUFFER_SIZE;
            this.ip = ip;
            this.connectionID = connectionID;
            if (connectionChanged != null)
                connectionChanged.Invoke(this, ConnectionState.open);
        }

        /// <summary>
        /// Starts this item packet processor
        /// MUST be called before sending data
        /// </summary>
        public void startPacketProcessing()
        {
            if (!isConnected)
            {
                this.isConnected = true;
                //Out.writeLine("Starting packet processsing of client [" + this.connectionID + "]", Out.logFlags.lowLogLevel);
                try
                {
                    this.dataSocket.BeginReceive(buffer, 0, buffer.Length, SocketFlags.None, new AsyncCallback(incomingDataPacket), dataSocket);
                }
                catch
                {
                    this.disconnect();
                }
            }
        }

        #endregion

        #region getters

        /// <summary>
        /// Returns the ip of the current connection
        /// </summary>
        /// <returns>The ip of this connection</returns>
        public string getIp()
        {
            return ip;
        }

        /// <summary>
        /// Returns the connection id
        /// </summary>
        /// <returns>The id of the connection</returns>
        public int getConnectionID()
        {
            return connectionID;
        }

        #endregion

        #region methods

        #region connection management
        /// <summary>
        /// Disconnects the current connection
        /// </summary>
        internal void disconnect()
        {
            if (isConnected)
            {
                this.isConnected = false;
                Out.writeLine("Connection [" + this.connectionID + "] has been disconnected", Out.logFlags.BelowStandardlogLevel);
                try
                {
                    if (this.dataSocket != null && this.dataSocket.Connected)
                    {
                        dataSocket.Close();
                    }
                }
                catch { }
                parser.Dispose();
                parser = null;

                try
                {
                    if (connectionChanged != null)
                        connectionChanged.Invoke(this, ConnectionState.closed);
                }
                catch (Exception Ex)
                {
                    Out.writeError(Ex.ToString(), Out.logFlags.ImportantLogLevel);

                    Out.writeSeriousError(Ex.ToString());
                }
                this.connectionChanged = null;
            }
            else
            {
                Out.writeLine("Connection [" + this.connectionID + "] has already been disconnected - ignoring disconnect call", Out.logFlags.BelowStandardlogLevel);
            }
        }


        /// <summary>
        /// Disposes the current item
        /// </summary>
        public void Dispose()
        {
            if (isConnected)
            {
                this.disconnect();
            }
        }
        #endregion

        #region data receiving

        /// <summary>
        /// Receives a packet of data and processes it
        /// </summary>
        /// <param name="iAr">The interface of an async result</param>
        private void incomingDataPacket(IAsyncResult iAr)
        {
            Out.writeLine("Packet received from client [" + this.connectionID + "]", Out.logFlags.lowLogLevel);
            int bytesReceived;
            try
            {
                //The amount of bytes received in the packet
                bytesReceived = dataSocket.EndReceive(iAr);
            }
            #region socket error / closed
            catch (ObjectDisposedException) { disconnect(); return; }
            catch (SocketException ex)
            {
                Out.writeError("SocketException :\r\n" + ex.Message + "\r\n" + ex.StackTrace, Out.logFlags.lowLogLevel);
                disconnect(); return;
            }
            catch (Exception ex)
            {
                Out.writeError("Exception :\r\n" + ex.Message + "\r\n" + ex.StackTrace);
                disconnect();
                return;
            }
            if (bytesReceived == 0)
            {
                Out.writeLine("Socket send 0 bytes, asuming closure of socket", Out.logFlags.lowLogLevel);
                disconnect();
                return;
            }
            #endregion
            Out.writeLine("Packet size is [" + bytesReceived + "] from client [" + this.connectionID + "]", Out.logFlags.lowLogLevel);
            try
            {
                byte[] packet = new byte[bytesReceived];
                Array.Copy(buffer, packet, bytesReceived);
                handlePacketData(packet);
            }
            #region exception handeling and continueing receiving
            catch (PacketMalformedException ex)
            {
                Out.writeError("Packet was malformed: " + ex.Message);
                try
                {
                    ServerOutgoingPacket packet = new ServerOutgoingPacket(ClientOpCode.Generic_Error);
                    packet.WriteInt(0);
                    packet.WriteString(ex.Message);
                    this.sendData(packet);
                }
                catch { disconnect(); }
                if (this.connectionChanged != null)
                    connectionChanged.Invoke(this, ConnectionState.malfunctioning_packet);
            }

            catch (SocketException ex)
            {
                Out.writeError("SocketException :\r\n" + ex.Message + "\r\n" + ex.StackTrace, Out.logFlags.lowLogLevel);
                this.disconnect();
                return;
            }
            catch (SendableException ex)
            {
                if (ex.getPacket() != null)
                    this.sendData(ex.getPacket() as ServerOutgoingPacket);
            }
            catch (Exception ex)
            {
                Out.writeError("Error occured in the connection manager:\r\n" + ex.Message + "\r\n" + ex.StackTrace);
                this.disconnect();
                return;

            }

            finally
            {
                try
                {
                    Out.writeLine("Started listening for next message from client [" + this.connectionID + "]", Out.logFlags.lowLogLevel);

                    //and we keep looking for the next packet
                    this.dataSocket.BeginReceive(buffer, 0, buffer.Length, SocketFlags.None, new AsyncCallback(incomingDataPacket), dataSocket);
                }
                catch (SocketException)
                {
                    Out.writeLine("Socket exception while listening for next packet for client [" + this.connectionID + "]", Out.logFlags.lowLogLevel);
                    disconnect();
                }
                catch (ObjectDisposedException) { disconnect(); }
            }
            #endregion
        }


        /// <summary>
        /// Handles packet data
        /// </summary>
        /// <param name="packet">The data received by the </param>
        private void handlePacketData(byte[] packet)
        {
            if (parser != null)
                parser.handlePacketData(packet);
            //Check if the length is bigger than a single integer value

        }

        #endregion

        #region data sending
        /// <summary>
        /// Sends data to this instance
        /// </summary>
        /// <param name="Data">The data to be send</param>
        public void sendData(ServerOutgoingPacket data)
        {
            try
            {
                Out.writeLine("Sending packet [" + data.getOpCode().ToString() + "] to client [" + this.connectionID + "]", Out.logFlags.BelowStandardlogLevel);
                byte[] dataBytes = data.GetPacketData();
                this.dataSocket.BeginSend(dataBytes, 0, dataBytes.Length, 0, new AsyncCallback(sentData), this.dataSocket);
                
            }
            catch
            {
                disconnect();
            }
        }

        public void sendbData(byte[] packet)
        {
            sendData(packet);
        }
        private void sendData(byte[] packet)
        {
            Out.writeLine("Sending byte packet of length [" + packet.Length + "] to client [" + this.connectionID + "]", Out.logFlags.BelowStandardlogLevel);
            this.dataSocket.BeginSend(packet, 0, packet.Length, 0, new AsyncCallback(sentData), this.dataSocket);
        }

        /// <summary>
        /// Same as sendData
        /// </summary>
        /// <param name="iAr">The a-synchronious interface</param>
        private void sentData(IAsyncResult iAr)
        {
            try { 
                dataSocket.EndSend(iAr); 
            }
            catch { disconnect(); }
        }
        #endregion

        #endregion
    }
}
