﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Timers;
using SharedPacketLib.DataPackets;
using System.Diagnostics;
using Helpers;
using SocketTest.Packets;

namespace SocketTest.Managers
{
    public class PingChecker : IDisposable
    {

        /// <summary>
        /// Contains the timer to the ping interval
        /// </summary>
        private Timer myPingTimer;
        /// <summary>
        /// Contains the game connection handeler
        /// </summary>
        private ConnectionHandeler handeler;
        /// <summary>
        /// Indication if this item has send a ping request
        /// </summary>
        private bool sendPing = false;
        /// <summary>
        /// Indication if the client has responded with the pong response
        /// </summary>
        private bool receivedPing = false;

        /// <summary>
        /// Contains the current ping response time
        /// </summary>
        private Stopwatch pingResponseTime;

        /// <summary>
        /// The ping of this item
        /// </summary>
        private long lastResponseTime = 9999;
#if DEBUG
        ~PingChecker()
        {
            Out.writeSpecialLine("Disposed PingChecker ", Out.logFlags.lowLogLevel, ConsoleColor.Red, ConsoleColor.Yellow, "Disposer -> ");
        }
#endif
        /// <summary>
        /// This method initializes a ping checker for the given connection
        /// </summary>
        /// <param name="handeler">The connection to poll</param>
        public PingChecker(ConnectionHandeler handeler)
        {
            this.handeler = handeler;
            pingResponseTime = new Stopwatch();

            int pingInterval = AwsomeLicense.Configuration.getInt("ping.interval");
            
            handeler.Disconnection += userDisconnected;
            if (pingInterval <= 1)
                pingInterval = 1;
            pingInterval *= 1000;
            myPingTimer = new Timer(pingInterval);
            handeler.addPacketHandeler(ServerOpCode.Ping_Response, handleReceivedPing);
            myPingTimer.Elapsed += new ElapsedEventHandler(doPingCheck);
            myPingTimer.Start();
        }

        /// <summary>
        /// Called when the user disconnects
        /// </summary>
        private void userDisconnected(ConnectionHandeler handeler)
        {
            myPingTimer.Elapsed -= doPingCheck;
            handeler.Disconnection += userDisconnected;
            
            handeler.removeHandeler(ServerOpCode.Ping_Response);
            myPingTimer.Dispose();
        }

        /// <summary>
        /// Does a ping check
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void doPingCheck(object sender, ElapsedEventArgs e)
        {
            if (sendPing && receivedPing)
            {
                sendPing = false;
                receivedPing = false;
            }
            else if (sendPing && !receivedPing)
            {
                Out.writeError("Connection with id "+this.handeler.ID+" has been disconnected, Took to long to respond", Out.logFlags.BelowStandardlogLevel);
                handeler.Dispose();
            }
            else if (!sendPing)
            {
                sendNewPing();
            }
        }


        /// <summary>
        /// Sends a new ping to the client
        /// </summary>
        private void sendNewPing()
        {
            Out.writeNotification("Sending ping to connection " + handeler.ID);
            this.sendPing = true;
            this.receivedPing = false;
            this.pingResponseTime.Restart();

            this.handeler.sendPacket(new PingRequestPacket());
        }

        /// <summary>
        /// Sends a pingOK packet to the client and handles the received ping
        /// </summary>
        /// <param name="packet"></param>
        private void handleReceivedPing(ServerIncomingPacket packet)
        {
            if (sendPing && !receivedPing)
            {
                this.pingResponseTime.Stop();
                lastResponseTime = this.pingResponseTime.ElapsedMilliseconds;
                Out.writeNotification(string.Format("Connection {0} ping response: ping = {1}ms", handeler.ID, pingResponseTime.ElapsedMilliseconds));
                this.receivedPing = true;
            }
            handeler.sendPacket(new PingOkPacket());
        }

        public void Dispose()
        {
            if(myPingTimer != null)
                myPingTimer.Dispose();
            sendPing = false;
            receivedPing = false;

        }

        internal long getPing()
        {
            return this.lastResponseTime;
        }
    }
}

